/*
 * echo-client.c -- Echo client application for Pthreads port of dccp-tp
 *
 * Copyright (C) 2008 Tom Phelan
 *
 * This file is part of dccp-tp.
 *
 * Dccp-tp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Dccp-tp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with dccp-tp.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Documentation and source code for dccp-tp is available at
 * http://www.phelan-4.com/dccp-tp/.
 */

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "dctpCore.h"
#include "dctpApi.h"
#include "dctpSupport.h"

#define ECHO_DEFPORT 7
#define ECHO_MAXPKTS 300
#define ECHO_PKTSIZE 1400

static uint16_t echoPort = ECHO_DEFPORT;
static uint_t maxpkts = ECHO_MAXPKTS;
static uint_t pktsize = ECHO_PKTSIZE;
static uint8_t xbuf[ECHO_PKTSIZE];
static uint8_t rbuf[ECHO_PKTSIZE];

static int running;

static void *readThread(void *arg) {
    int sock = (int)arg;
    int rpkts = 0;

    while (running) {
	if (dccpRecv(sock, rbuf, sizeof(rbuf)) < 0) {
	    dctpoLog(DCTPLOG_ERR, "echoClientReadThread: can't recieve: %s\n", strerror(errno));
	    break;
	}
	++rpkts;
	/*dctpoLog(DCTPLOG_DEBUG, "echoClientReadThread: received packet %d\n", rpkts);*/
    }
    dctpoLog(DCTPLOG_INFO, "echoClientReadThread: exiting, received %d packets\n", rpkts);
    return(NULL);
}

static void *echoClientThread(void *arg) {
    int sock;
    uint32_t pktcnt;
    dctpSockaddrV4 addr;
    pthread_attr_t pattr;
    pthread_t thread;
    static struct timespec pause = {0, 1};

    for (sock = 0; sock < ECHO_PKTSIZE; ++sock) {
	xbuf[sock] = sock;
    }
    while (1) {
	dctpoLog(DCTPLOG_INFO, "echoClientThread: attempting connection...\n");
	if ((sock = dccpSocket(DCTP_PF_INET, SOCK_DCCP_NAT, "SC:ECHO")) < 0) {
	    dctpoLog(DCTPLOG_ERR, "echoClientThread: can't open socket: %d\n", errno);
	    return(NULL);
	}
	addr.sin_family = DCTP_AF_INET;
	addr.sin_port = dctpoHtons(echoPort);
	addr.sin_addr.s_addr = dctpoHtonl(DCTP_INADDR_LOOPBACK);
	if (dccpConnect(sock, (dctpSockaddr *)&addr, sizeof(addr), NULL, 0) < 0) {
	    dctpoLog(DCTPLOG_ERR, "echoClientThread: can't connect: %d\n", errno);
	    dccpClose(sock);
	    return(NULL);
	}
	dctpoLog(DCTPLOG_INFO, "echoClientThread: connected to %s\n",
		 inet_ntoa(*((struct in_addr *)&(addr.sin_addr))));
	running = 1;
	pthread_attr_init(&pattr);
	pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);
	pthread_create(&thread, &pattr, readThread, (void *)sock);
	sleep(1);
	dctpoLog(DCTPLOG_DEBUG, "echoClientThread: starting transmission\n");
	for (pktcnt = 0; pktcnt < maxpkts; ++pktcnt) {
	    *((uint32_t *)xbuf) = pktcnt;
	    if ((pktcnt % 100) == 0) {
		printf("echoClient: sending packet %d\r", pktcnt);
		fflush(stdout);
	    }
	    if (dccpSend(sock, xbuf, pktsize) < 0) {
		dctpoLog(DCTPLOG_ERR, "echoClientThread: can't send: %s\n", strerror(errno));
		break;
	    }
	    nanosleep(&pause, NULL);
	}
	printf("\n");
	sleep(1);
	dctpoLog(DCTPLOG_DEBUG, "echo-client: stopping\n");
	running = 0;
	sleep(1);
	dctpoLog(DCTPLOG_DEBUG, "echo-client: closing\n");
	dccpClose(sock);
	dctpoLog(DCTPLOG_DEBUG, "echoClient: sent %d packets, closing\n", pktcnt);
	sleep(3);
	dctpoLog(DCTPLOG_DEBUG, "echoClient: restarting\n\n\n");
    }
    return(NULL);
}

void echoClientInit(void) {
    pthread_attr_t pattr;
    pthread_t thread;

    pthread_attr_init(&pattr);
    pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);
    pthread_create(&thread, &pattr, echoClientThread, NULL);
}
